import sys
#In package unittest there is no module mock before python 3.3. In python 3
#module mock is not imported by * because module mock is not imported in
#unittest/__init__.py
major = sys.version_info.major
from unittest import *
try:
    import unittest.mock as mock
except ImportError:
    import mock

#backport of assert_not_called (new in version 3.5)
if not hasattr(mock.Mock, "assert_not_called"):
    def __assert_not_called(self, *args, **kwargs):
        if self.call_count != 0:
            msg = ("Expected '%s' to not have been called. Called %s times." %
                   (self._mock_name or 'mock', self.call_count))
            raise AssertionError(msg)
    mock.Mock.assert_not_called = __assert_not_called


if not hasattr(mock, "mock_open"):
    def create_mock_open(MagicMock, DEFAULT, inPy3k):
        """
        Backport mock_open for older mock versions.
        Backport is taken from mock package. Original code is slightly adapted:
         * code is covered by create_mock_open
         * MagicMock, DEFAULT, and inPy3k flag was originally module globals
         * file_spec was originally plain value (and in mock_open referenced as
           global) but we want keep namespace of this module as clean as
           possible so now is file_spec encapsulated.
        """
        file_spec = [None]

        def _iterate_read_data(read_data):
            # Helper for mock_open:
            # Retrieve lines from read_data via a generator so that separate
            # calls to readline, read, and readlines are properly interleaved
            data_as_list = ['{0}\n'.format(l) for l in read_data.split('\n')]

            if data_as_list[-1] == '\n':
                # If the last line ended in a newline, the list comprehension
                #will #have an extra entry that's just a newline.  Remove this.
                data_as_list = data_as_list[:-1]
            else:
                # If there wasn't an extra newline by itself, then the file
                # being emulated doesn't have a newline to end the last line
                # remove the newline that our naive format() added
                data_as_list[-1] = data_as_list[-1][:-1]

            for line in data_as_list:
                yield line

        def mock_open(mock=None, read_data=''):
            """
            A helper function to create a mock to replace the use of `open`. It
            works for `open` called directly or used as a context manager.

            The `mock` argument is the mock object to configure. If `None` (the
            default) then a `MagicMock` will be created for you, with the API
            limited to methods or attributes available on standard file
            handles.

            `read_data` is a string for the `read` methoddline`, and
            `readlines` of the file handle to return.  This is an empty string
            by default.
            """
            def _readlines_side_effect(*args, **kwargs):
                if handle.readlines.return_value is not None:
                    return handle.readlines.return_value
                return list(_state[0])

            def _read_side_effect(*args, **kwargs):
                if handle.read.return_value is not None:
                    return handle.read.return_value
                return ''.join(_state[0])

            def _readline_side_effect():
                if handle.readline.return_value is not None:
                    while True:
                        yield handle.readline.return_value
                for line in _state[0]:
                    yield line

            if file_spec[0] is None:
                # set on first use
                if inPy3k:
                    import _io
                    file_spec[0] = list(
                        set(dir(_io.TextIOWrapper))
                            .union(set(dir(_io.BytesIO)))
                    )
                else:
                    file_spec[0] = file

            if mock is None:
                mock = MagicMock(name='open', spec=open)

            handle = MagicMock(spec=file_spec[0])
            handle.__enter__.return_value = handle

            _state = [_iterate_read_data(read_data), None]

            handle.write.return_value = None
            handle.read.return_value = None
            handle.readline.return_value = None
            handle.readlines.return_value = None

            handle.read.side_effect = _read_side_effect
            _state[1] = _readline_side_effect()
            handle.readline.side_effect = _state[1]
            handle.readlines.side_effect = _readlines_side_effect

            def reset_data(*args, **kwargs):
                _state[0] = _iterate_read_data(read_data)
                if handle.readline.side_effect == _state[1]:
                    # Only reset the side effect if the user hasn't overridden
                    #it.
                    _state[1] = _readline_side_effect()
                    handle.readline.side_effect = _state[1]
                return DEFAULT

            mock.side_effect = reset_data
            mock.return_value = handle
            return mock
        return mock_open

    mock.mock_open = create_mock_open(
        mock.MagicMock,
        mock.DEFAULT,
        inPy3k=(major==3),
    )
    del create_mock_open

def ensure_raise_from_iterable_side_effect():
    """
    Adjust mock.Mock to raise when side_effect is iterable and efect item
    applied in the current call is exception (class or instance) and this
    exception is simply returned (in older version of mock).
    """
    def create_new_call(old_call, inPy3k):
        # pylint: disable=old-style-class
        class OldStyleClass:
            pass
        ClassTypes = (type,) if inPy3k else (type, type(OldStyleClass))

        def is_exception(obj):
            return isinstance(obj, BaseException) or (
                isinstance(obj, ClassTypes)
                and
                issubclass(obj, BaseException)
            )

        def new_call(_mock_self, *args, **kwargs):
            """
            Wrap original call.
            If side_effect is itterable and result is an exception then we
            raise this exception. Newer versions of mock it makes itself (so
            in this case exception is raised from old_call) but we need it
            for the old versions as well.
            """
            call_result = old_call(_mock_self, *args, **kwargs)
            try:
                iter(_mock_self.side_effect)
            except TypeError:
                return call_result

            if is_exception(call_result):
                raise call_result
            return call_result

        return new_call
    mock.Mock.__call__ = create_new_call(mock.Mock.__call__, inPy3k=(major==3))
ensure_raise_from_iterable_side_effect()

del major, sys



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from functools import partial
from pcs.lib.cib.constraint import order
import pcs.lib.commands.constraint.common

#configure common constraint command
show = partial(
    pcs.lib.commands.constraint.common.show,
    order.TAG_NAME,
    lambda element: element.attrib.has_key('first')
)

#configure common constraint command
create_with_set = partial(
    pcs.lib.commands.constraint.common.create_with_set,
    order.TAG_NAME,
    order.prepare_options_with_set
)



from __future__ import (
    absolute_import,
    division,
    print_function,
)

import logging

from pcs.test.tools.pcs_unittest import TestCase

from pcs.test.tools.pcs_unittest import mock
from pcs.test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
)
from pcs.test.tools.custom_mock import MockLibraryReportProcessor
from pcs.test.tools.integration_lib import Runner
from pcs.test.tools.command_env.mock_runner import Call

from pcs import settings
from pcs.common import report_codes
from pcs.common.node_communicator import RequestTarget
from pcs.lib.errors import (
    ReportItemSeverity as Severities,
    LibraryError,
    ReportItem,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.node import (
    NodeAddresses,
    NodeAddressesList,
)
import pcs.lib.commands.sbd as cmd_sbd


runner = Runner()


def _assert_equal_list_of_dictionaries_without_order(expected, actual):
    for item in actual:
        if item not in expected:
            raise AssertionError("Given but not expected: {0}".format(item))
    for item in expected:
        if item not in actual:
            raise AssertionError("Expected but not given: {0}".format(item))


class ValidateSbdOptionsTest(TestCase):
    def setUp(self):
        self.allowed_sbd_options = sorted([
            "SBD_DELAY_START", "SBD_STARTMODE", "SBD_WATCHDOG_TIMEOUT"
        ])

    def test_all_ok(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean"
        }
        self.assertEqual([], cmd_sbd._validate_sbd_options(config))

    def test_unknown_options(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_UNKNOWN": "",
            "another_unknown_option": "some value"
        }
        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_UNKNOWN"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    report_codes.FORCE_OPTIONS
                ),
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["another_unknown_option"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    report_codes.FORCE_OPTIONS
                )
            ]
        )

    def test_unknown_options_forced(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_UNKNOWN": "",
            "another_unknown_option": "some value"
        }
        # just make sure there is no exception raised
        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config, allow_unknown_opts=True),
            [
                (
                    Severities.WARNING,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_UNKNOWN"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.WARNING,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["another_unknown_option"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                )
            ]
        )


    def test_unsupported_options(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_WATCHDOG_DEV": "/dev/watchdog",
            "SBD_OPTS": "  ",
            "SBD_DEVICE": "/dev/vda",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_WATCHDOG_DEV"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_OPTS"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_DEVICE"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                )
            ]
        )

    def test_invalid_and_unsupported_options(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_WATCHDOG_DEV": "/dev/watchdog",
            "SBD_UNKNOWN": "",
            "SBD_OPTS": "  "
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_WATCHDOG_DEV"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_OPTS"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_UNKNOWN"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    report_codes.FORCE_OPTIONS
                )
            ]
        )

    def test_invalid_and_unsupported_options_forced(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_WATCHDOG_DEV": "/dev/watchdog",
            "SBD_UNKNOWN": "",
            "SBD_OPTS": "  ",
            "SBD_PACEMAKER": "false",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config, allow_unknown_opts=True),
            [
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_WATCHDOG_DEV"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_OPTS"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.WARNING,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_UNKNOWN"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                ),
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["SBD_PACEMAKER"],
                        "option_type": None,
                        "allowed": self.allowed_sbd_options,
                        "allowed_patterns": [],
                    },
                    None
                )
            ]
        )

    def test_watchdog_timeout_is_not_present(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_STARTMODE": "clean"
        }
        self.assertEqual([], cmd_sbd._validate_sbd_options(config))

    def test_watchdog_timeout_is_nonnegative_int(self):
        config = {
            "SBD_WATCHDOG_TIMEOUT": "-1",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTION_VALUE,
                    {
                        "option_name": "SBD_WATCHDOG_TIMEOUT",
                        "option_value": "-1",
                        "allowed_values": "a non-negative integer",
                    },
                    None
                )
            ]
        )

    def test_watchdog_timeout_is_not_int(self):
        config = {
            "SBD_WATCHDOG_TIMEOUT": "not int",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTION_VALUE,
                    {
                        "option_name": "SBD_WATCHDOG_TIMEOUT",
                        "option_value": "not int",
                        "allowed_values": "a non-negative integer",
                    },
                    None
                )
            ]
        )

    def test_watchdog_timeout_is_none(self):
        config = {
            "SBD_WATCHDOG_TIMEOUT": None,
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                (
                    Severities.ERROR,
                    report_codes.INVALID_OPTION_VALUE,
                    {
                        "option_name": "SBD_WATCHDOG_TIMEOUT",
                        "option_value": None,
                        "allowed_values": "a non-negative integer",
                    },
                    None
                )
            ]
        )


class ValidateWatchdogDictTest(TestCase):
    def test_all_ok(self):
        watchdog_dict = {
            NodeAddresses("node1"): "/dev/watchdog1",
            NodeAddresses("node2"): "/dev/watchdog2",
        }
        self.assertEqual([], cmd_sbd._validate_watchdog_dict(watchdog_dict))

    def test_some_not_ok(self):
        watchdog_dict = {
            NodeAddresses("node1"): "",
            NodeAddresses("node2"): None,
            NodeAddresses("node3"): "/dev/watchdog",
            NodeAddresses("node4"): "../dev/watchdog",
        }
        assert_report_item_list_equal(
            cmd_sbd._validate_watchdog_dict(watchdog_dict),
            [
                (
                    Severities.ERROR,
                    report_codes.WATCHDOG_INVALID,
                    {"watchdog": watchdog}
                ) for watchdog in ["", None, "../dev/watchdog"]
            ]
        )


class ValidateDeviceDictTest(TestCase):
    def test_all_ok(self):
        device_dict = {
            "node1": ["/dev1", "/dev2"],
            "node2": ["/dev1"],
        }
        self.assertEqual([], cmd_sbd._validate_device_dict(device_dict))

    def test_some_not_ok(self):
        too_many_devices = [
            "dev" + str(i) for i in range(settings.sbd_max_device_num + 1)
        ]
        device_dict = {
            "node1": [],
            "node2": too_many_devices,
            "node3": ["/dev/vda"],
            "node4": ["/dev/vda1", "../dev/sda2"],
        }
        assert_report_item_list_equal(
            cmd_sbd._validate_device_dict(device_dict),
            [
                (
                    Severities.ERROR,
                    report_codes.SBD_NO_DEVICE_FOR_NODE,
                    {
                        "node": "node1",
                    }
                ),
                (
                    Severities.ERROR,
                    report_codes.SBD_TOO_MANY_DEVICES_FOR_NODE,
                    {
                        "node": "node2",
                        "device_list": too_many_devices,
                        "max_devices": settings.sbd_max_device_num,
                    }
                ),
                (
                    Severities.ERROR,
                    report_codes.SBD_DEVICE_PATH_NOT_ABSOLUTE,
                    {
                        "node": "node4",
                        "device": "../dev/sda2",
                    }
                ),
            ]
        )


class CheckNodeNamesInClusterTest(TestCase):
    def setUp(self):
        self.node_list = NodeAddressesList([
            NodeAddresses("node1"),
            NodeAddresses("node2"),
            NodeAddresses("node3"),
        ])

    def test_all_ok(self):
        node_name_list = ["node1", "node3"]
        self.assertEqual(
            [],
            cmd_sbd._check_node_names_in_cluster(self.node_list, node_name_list)
        )

    def test_repeating_nodes(self):
        node_name_list = ["node1", "node3", "node1"]
        self.assertEqual(
            [],
            cmd_sbd._check_node_names_in_cluster(self.node_list, node_name_list)
        )

    def test_one_not_found(self):
        node_name_list = ["node0", "node3"]
        assert_report_item_list_equal(
            cmd_sbd._check_node_names_in_cluster(
                self.node_list, node_name_list
            ),
            [(
                Severities.ERROR,
                report_codes.NODE_NOT_FOUND,
                {"node": "node0"}
            )]
        )

    def test_multiple_not_found(self):
        node_name_list = ["node0", "node3", "node4"]
        assert_report_item_list_equal(
            cmd_sbd._check_node_names_in_cluster(
                self.node_list, node_name_list
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.NODE_NOT_FOUND,
                    {"node": node}
                ) for node in ["node0", "node4"]
            ]
        )

    def test_multiple_not_found_repeating(self):
        node_name_list = ["node0", "node3", "node4", "node0", "node1", "node3"]
        assert_report_item_list_equal(
            cmd_sbd._check_node_names_in_cluster(
                self.node_list, node_name_list
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.NODE_NOT_FOUND,
                    {"node": node}
                ) for node in ["node0", "node4"]
            ]
        )


class GetFullTargetDictTest(TestCase):
    def setUp(self):
        self.target_list = [
            RequestTarget("node{0}".format(i)) for i in range(1, 4)
        ]

    def test_not_using_default(self):
        target_dict = dict([
            ("node" + str(i), "val" + str(i)) for i in range(4)
        ])
        expected = {
            self.target_list[0].label: "val1",
            self.target_list[1].label: "val2",
            self.target_list[2].label: "val3",
        }
        self.assertEqual(
            expected,
            cmd_sbd._get_full_target_dict(self.target_list, target_dict, None)
        )

    def test_using_default(self):
        target_dict = dict([
            ("node" + str(i), "val" + str(i)) for i in range(3)
        ])
        default = "default"
        expected = {
            self.target_list[0].label: "val1",
            self.target_list[1].label: "val2",
            self.target_list[2].label: default,
        }
        self.assertEqual(
            expected,
            cmd_sbd._get_full_target_dict(
                self.target_list, target_dict, default
            )
        )


@mock.patch("pcs.lib.sbd.get_local_sbd_config")
class GetLocalSbdConfigTest(TestCase):
    def setUp(self):
        self.mock_env = mock.MagicMock(spec_set=LibraryEnvironment)

    def test_success(self, mock_config):
        self.mock_env.is_cman_cluster = False
        mock_config.return_value = """
# This file has been generated by pcs.
SBD_OPTS="-n node1"
SBD_WATCHDOG_DEV=/dev/watchdog
SBD_WATCHDOG_TIMEOUT=0
"""

        self.assertEqual(
            {
                "SBD_OPTS": '"-n node1"',
                "SBD_WATCHDOG_DEV": "/dev/watchdog",
                "SBD_WATCHDOG_TIMEOUT": "0"
            },
            cmd_sbd.get_local_sbd_config(self.mock_env)
        )
        self.assertEqual(1, mock_config.call_count)

    def test_file_error(self, mock_config):
        self.mock_env.is_cman_cluster = False
        mock_config.side_effect = LibraryError(ReportItem.error(
            report_codes.UNABLE_TO_GET_SBD_CONFIG,
        ))
        assert_raise_library_error(
            lambda: cmd_sbd.get_local_sbd_config(self.mock_env),
            (
                Severities.ERROR,
                report_codes.UNABLE_TO_GET_SBD_CONFIG,
                {}
            )
        )


class CommonTest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.patcher = mock.patch.object(
            LibraryEnvironment,
            "cmd_runner",
            lambda self: runner
        )
        cls.patcher.start()

    @classmethod
    def tearDownClass(cls):
        cls.patcher.stop()

    def setUp(self):
        self.env = LibraryEnvironment(
            mock.MagicMock(logging.Logger),
            MockLibraryReportProcessor()
        )


class InitializeBlockDevicesTest(CommonTest):
    def fixture_sbd_init(
        self, device_list, options, stdout="", stderr="", return_code=0
    ):
        cmd = ["sbd"]
        for device in device_list:
            cmd += ["-d", device]

        for opt, val in options:
            cmd += [opt, val]

        cmd.append("create")
        return [Call(" ".join(cmd), stdout, stderr, return_code)]

    def fixture_invalid_value(self, option, value):
        return (
            Severities.ERROR,
            report_codes.INVALID_OPTION_VALUE,
            {
                "option_name": option,
                "option_value": value,
                "allowed_values": "a non-negative integer",
            }
        )

    def test_all_options(self):
        device_list = ["dev1", "dev2"]
        option_dict = {
            "watchdog-timeout": "1",
            "loop-timeout": "10",
            "allocate-timeout": "3",
            "msgwait-timeout": "2",

        }
        runner.set_runs(self.fixture_sbd_init(
            device_list,
            [
                ("-2", "3"),
                ("-3", "10"),
                ("-4", "2"),
                ("-1", "1"),
            ]
        ))
        cmd_sbd.initialize_block_devices(self.env, device_list, option_dict)
        runner.assert_everything_launched()
        self.env.report_processor.assert_reports([
            (
                Severities.INFO,
                report_codes.SBD_DEVICE_INITIALIZATION_STARTED,
                {"device_list": device_list}
            ),
            (
                Severities.INFO,
                report_codes.SBD_DEVICE_INITIALIZATION_SUCCESS,
                {"device_list": device_list}
            ),
        ])

    def test_no_options(self):
        device_list = ["dev1", "dev2"]
        runner.set_runs(self.fixture_sbd_init(device_list,{}))
        cmd_sbd.initialize_block_devices(self.env, device_list, {})
        runner.assert_everything_launched()
        self.env.report_processor.assert_reports([
            (
                Severities.INFO,
                report_codes.SBD_DEVICE_INITIALIZATION_STARTED,
                {"device_list": device_list}
            ),
            (
                Severities.INFO,
                report_codes.SBD_DEVICE_INITIALIZATION_SUCCESS,
                {"device_list": device_list}
            ),
        ])

    def test_validation_failed(self):
        option_dict = {
            "unknown_option": "val",
            "watchdog-timeout": "-1",
            "another_one": "-1",
            "loop-timeout": "-3",
            "allocate-timeout": "-3",
            "msgwait-timeout": "-2",
        }
        allowed_options = [
            "watchdog-timeout", "loop-timeout", "allocate-timeout",
            "msgwait-timeout",
        ]
        assert_raise_library_error(
            lambda: cmd_sbd.initialize_block_devices(self.env, [], option_dict),
            (
                Severities.ERROR,
                report_codes.REQUIRED_OPTION_IS_MISSING,
                {
                    "option_names": ["device"],
                    "option_type": None,
                }
            ),
            (
                Severities.ERROR,
                report_codes.INVALID_OPTIONS,
                {
                    "option_names": sorted(["another_one", "unknown_option"]),
                    "option_type": "option",
                    "allowed": sorted(allowed_options),
                    "allowed_patterns": [],
                }
            ),
            *[
                self.fixture_invalid_value(opt, option_dict[opt])
                for opt in allowed_options
            ]
        )


@mock.patch("os.path.exists")
@mock.patch("pcs.lib.sbd.get_local_sbd_config")
@mock.patch("pcs.lib.external.is_systemctl", lambda: True)
class GetLocalDevicesInfoTest(CommonTest):
    def fixture_sbd_enabled(self, enabled):
        cmd = [settings.systemctl_binary, "is-enabled", "sbd.service"]
        return [Call(" ".join(cmd), returncode=0 if enabled else 1)]

    def fixture_sbd_info(self, device, stdout="", return_code=0):
        cmd = ["sbd", "-d", device, "list"]
        return [Call(" ".join(cmd), stdout, returncode=return_code)]

    def fixture_sbd_dump(self, device, stdout="", return_code=0):
        cmd = ["sbd", "-d", device, "dump"]
        return [Call(" ".join(cmd), stdout, returncode=return_code)]

    def test_success(self, mock_config, mock_config_exists):
        mock_config_exists.return_value = True
        mock_config.return_value = """
SBD_DEVICE="/dev1;/dev2"
        """
        runner.set_runs(
            self.fixture_sbd_enabled(True) +
            self.fixture_sbd_info("/dev1", "1") +
            self.fixture_sbd_info("/dev2", "2")
        )
        expected_output = [
            {
                "device": "/dev1",
                "list": "1",
                "dump": None,
            },
            {
                "device": "/dev2",
                "list": "2",
                "dump": None,
            },
        ]
        self.assertEqual(
            expected_output, cmd_sbd.get_local_devices_info(self.env)
        )
        runner.assert_everything_launched()
        mock_config.assert_called_once_with()
        mock_config_exists.assert_called_once_with(settings.sbd_config)

    def test_with_dump(self, mock_config, mock_config_exists):
        mock_config_exists.return_value = True
        mock_config.return_value = """
SBD_DEVICE="/dev1;/dev2"
        """
        runner.set_runs(
            self.fixture_sbd_enabled(True) +
            self.fixture_sbd_info("/dev1", "1") +
            self.fixture_sbd_dump("/dev1", "3") +
            self.fixture_sbd_info("/dev2", "2") +
            self.fixture_sbd_dump("/dev2", "4")
        )
        expected_output = [
            {
                "device": "/dev1",
                "list": "1",
                "dump": "3",
            },
            {
                "device": "/dev2",
                "list": "2",
                "dump": "4",
            },
        ]
        self.assertEqual(
            expected_output, cmd_sbd.get_local_devices_info(self.env, dump=True)
        )
        runner.assert_everything_launched()
        mock_config.assert_called_once_with()
        mock_config_exists.assert_called_once_with(settings.sbd_config)

    def test_no_config(self, mock_config, mock_config_exists):
        mock_config_exists.return_value = False
        runner.set_runs(self.fixture_sbd_enabled(True))
        self.assertEqual([], cmd_sbd.get_local_devices_info(self.env))
        runner.assert_everything_launched()
        self.assertEqual(0, mock_config.call_count)
        mock_config_exists.assert_called_once_with(settings.sbd_config)

    def test_sbd_disabled(self, mock_config, mock_config_exists):
        mock_config_exists.return_value = True
        runner.set_runs(self.fixture_sbd_enabled(False))
        self.assertEqual([], cmd_sbd.get_local_devices_info(self.env))
        runner.assert_everything_launched()
        self.assertEqual(0, mock_config.call_count)
        self.assertEqual(0, mock_config_exists.call_count)

    def test_with_failures(self, mock_config, mock_config_exists):
        mock_config_exists.return_value = True
        mock_config.return_value = """
SBD_DEVICE="/dev1;/dev2;/dev3"
        """
        runner.set_runs(
            self.fixture_sbd_enabled(True) +
            self.fixture_sbd_info("/dev1", "1", 1) +
            self.fixture_sbd_info("/dev2", "2") +
            self.fixture_sbd_dump("/dev2", "4", 1) +
            self.fixture_sbd_info("/dev3", "5") +
            self.fixture_sbd_dump("/dev3", "6")
        )
        expected_output = [
            {
                "device": "/dev1",
                "list": None,
                "dump": None,
            },
            {
                "device": "/dev2",
                "list": "2",
                "dump": None,
            },
            {
                "device": "/dev3",
                "list": "5",
                "dump": "6",
            },
        ]
        self.assertEqual(
            expected_output, cmd_sbd.get_local_devices_info(self.env, dump=True)
        )
        self.env.report_processor.assert_reports([
            (
                Severities.WARNING,
                report_codes.SBD_DEVICE_LIST_ERROR,
                {
                    "device": "/dev1",
                    "reason": "1"
                }
            ),
            (
                Severities.WARNING,
                report_codes.SBD_DEVICE_DUMP_ERROR,
                {
                    "device": "/dev2",
                    "reason": "4"
                }
            ),
        ])
        runner.assert_everything_launched()
        mock_config.assert_called_once_with()
        mock_config_exists.assert_called_once_with(settings.sbd_config)


class SetMessageTest(CommonTest):
    def fixture_call_sbd_message(
        self, device, node, message, stderr="", return_code=0
    ):
        return [Call(
            "sbd -d {0} message {1} {2}".format(device, node, message),
            stderr=stderr,
            returncode=return_code
        )]

    def test_empty_options(self):
        assert_raise_library_error(
            lambda: cmd_sbd.set_message(self.env, "", "", ""),
            (
                Severities.ERROR,
                report_codes.REQUIRED_OPTION_IS_MISSING,
                {
                    "option_names": ["device", "node"],
                    "option_type": None,
                }
            ),
            (
                Severities.ERROR,
                report_codes.INVALID_OPTION_VALUE,
                {
                    "option_name": "message",
                    "option_value": "",
                    "allowed_values": settings.sbd_message_types,
                }
            )
        )

    def test_invalid_message_type(self):
        assert_raise_library_error(
            lambda: cmd_sbd.set_message(self.env, "device", "node1", "message"),
            (
                Severities.ERROR,
                report_codes.INVALID_OPTION_VALUE,
                {
                    "option_name": "message",
                    "option_value": "message",
                    "allowed_values": settings.sbd_message_types,
                }
            )
        )

    def test_success(self):
        runner.set_runs(self.fixture_call_sbd_message("device", "node", "test"))
        cmd_sbd.set_message(self.env, "device", "node", "test")
        runner.assert_everything_launched()

    def test_failuer(self):
        runner.set_runs(
            self.fixture_call_sbd_message("device", "node", "test", "error", 1)
        )
        assert_raise_library_error(
            lambda: cmd_sbd.set_message(self.env, "device", "node", "test"),
            (
                Severities.ERROR,
                report_codes.SBD_DEVICE_MESSAGE_ERROR,
                {
                    "device": "device",
                    "node": "node",
                    "message": "test",
                    "reason": "error",
                }
            )
        )
        runner.assert_everything_launched()



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.common import report_codes
from pcs.lib.commands import resource
from pcs.lib.errors import ReportItemSeverity as severities
from pcs.test.tools import fixture
from pcs.test.tools.command_env import get_env_tools
from pcs.test.tools.pcs_unittest import TestCase


fixture_primitive_cib_managed = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
        </primitive>
    </resources>
"""
fixture_primitive_cib_unmanaged = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        </primitive>
    </resources>
"""

fixture_primitive_cib_managed_op_enabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master" />
                <op id="A-monitor-s" name="monitor" role="Slave" />
            </operations>
        </primitive>
    </resources>
"""
fixture_primitive_cib_managed_op_disabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master"
                    enabled="false" />
                <op id="A-monitor-s" name="monitor" role="Slave"
                    enabled="false" />
            </operations>
        </primitive>
    </resources>
"""
fixture_primitive_cib_unmanaged_op_enabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master" />
                <op id="A-monitor-s" name="monitor" role="Slave" />
            </operations>
        </primitive>
    </resources>
"""
fixture_primitive_cib_unmanaged_op_disabled = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Stateful">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <operations>
                <op id="A-start" name="start" />
                <op id="A-stop" name="stop" />
                <op id="A-monitor-m" name="monitor" role="Master"
                    enabled="false" />
                <op id="A-monitor-s" name="monitor" role="Slave"
                    enabled="false" />
            </operations>
        </primitive>
    </resources>
"""

fixture_group_cib_managed = """
    <resources>
        <group id="A">
            <primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </group>
    </resources>
"""
fixture_group_cib_unmanaged_resource = """
    <resources>
        <group id="A">
            <primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A1-meta_attributes">
                    <nvpair id="A1-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
            <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </group>
    </resources>
"""
fixture_group_cib_unmanaged_resource_and_group = """
    <resources>
        <group id="A">
            <primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A1-meta_attributes">
                    <nvpair id="A1-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
            <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
        </group>
    </resources>
"""
fixture_group_cib_unmanaged_all_resources = """
    <resources>
        <group id="A">
            <primitive id="A1" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A1-meta_attributes">
                    <nvpair id="A1-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
            <primitive id="A2" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A2-meta_attributes">
                    <nvpair id="A2-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </group>
    </resources>
"""

fixture_clone_cib_managed = """
    <resources>
        <clone id="A-clone">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </clone>
    </resources>
"""
fixture_clone_cib_unmanaged_clone = """
    <resources>
        <clone id="A-clone">
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </clone>
    </resources>
"""
fixture_clone_cib_unmanaged_primitive = """
    <resources>
        <clone id="A-clone">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </clone>
    </resources>
"""
fixture_clone_cib_unmanaged_both = """
    <resources>
        <clone id="A-clone">
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </clone>
    </resources>
"""

fixture_clone_cib_managed_op_enabled = """
    <resources>
        <clone id="A-clone">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor"/>
                </operations>
            </primitive>
        </clone>
    </resources>
"""
fixture_clone_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <clone id="A-clone">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </clone>
    </resources>
"""

fixture_master_cib_managed = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_master = """
    <resources>
        <master id="A-master">
            <meta_attributes id="A-master-meta_attributes">
                <nvpair id="A-master-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_primitive = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_both = """
    <resources>
        <master id="A-master">
            <meta_attributes id="A-master-meta_attributes">
                <nvpair id="A-master-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </master>
    </resources>
"""

fixture_master_cib_managed_op_enabled = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor"/>
                </operations>
            </primitive>
        </master>
    </resources>
"""
fixture_master_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <master id="A-master">
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </master>
    </resources>
"""

fixture_clone_group_cib_managed = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_primitive = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A1-meta_attributes">
                        <nvpair id="A1-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_all_primitives = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A1-meta_attributes">
                        <nvpair id="A1-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A2-meta_attributes">
                        <nvpair id="A2-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_clone = """
    <resources>
        <clone id="A-clone">
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_everything = """
    <resources>
        <clone id="A-clone">
            <meta_attributes id="A-clone-meta_attributes">
                <nvpair id="A-clone-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <group id="A">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A1-meta_attributes">
                        <nvpair id="A1-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A2-meta_attributes">
                        <nvpair id="A2-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
            </group>
        </clone>
    </resources>
"""

fixture_clone_group_cib_managed_op_enabled = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <operations>
                        <op id="A1-start" name="start" />
                        <op id="A1-stop" name="stop" />
                        <op id="A1-monitor" name="monitor" />
                    </operations>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <operations>
                        <op id="A2-start" name="start" />
                        <op id="A2-stop" name="stop" />
                        <op id="A2-monitor" name="monitor" />
                    </operations>
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A1-meta_attributes">
                        <nvpair id="A1-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <operations>
                        <op id="A1-start" name="start" />
                        <op id="A1-stop" name="stop" />
                        <op id="A1-monitor" name="monitor" enabled="false" />
                    </operations>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <operations>
                        <op id="A2-start" name="start" />
                        <op id="A2-stop" name="stop" />
                        <op id="A2-monitor" name="monitor" />
                    </operations>
                </primitive>
            </group>
        </clone>
    </resources>
"""
fixture_clone_group_cib_unmanaged_all_primitives_op_disabled = """
    <resources>
        <clone id="A-clone">
            <group id="A">
                <primitive id="A1" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A1-meta_attributes">
                        <nvpair id="A1-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <operations>
                        <op id="A1-start" name="start" />
                        <op id="A1-stop" name="stop" />
                        <op id="A1-monitor" name="monitor" enabled="false" />
                    </operations>
                </primitive>
                <primitive id="A2" class="ocf" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="A2-meta_attributes">
                        <nvpair id="A2-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                    <operations>
                        <op id="A2-start" name="start" />
                        <op id="A2-stop" name="stop" />
                        <op id="A2-monitor" name="monitor" enabled="false" />
                    </operations>
                </primitive>
            </group>
        </clone>
    </resources>
"""


fixture_bundle_empty_cib_managed = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
        </bundle>
    </resources>
"""
fixture_bundle_empty_cib_unmanaged_bundle = """
    <resources>
        <bundle id="A-bundle">
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <docker image="pcs:test" />
        </bundle>
    </resources>
"""

fixture_bundle_cib_managed = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </bundle>
    </resources>
"""
fixture_bundle_cib_unmanaged_bundle = """
    <resources>
        <bundle id="A-bundle">
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
            </primitive>
        </bundle>
    </resources>
"""
fixture_bundle_cib_unmanaged_primitive = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </bundle>
    </resources>
"""
fixture_bundle_cib_unmanaged_both = """
    <resources>
        <bundle id="A-bundle">
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </bundle>
    </resources>
"""

fixture_bundle_cib_managed_op_enabled = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor"/>
                </operations>
            </primitive>
        </bundle>
    </resources>
"""
fixture_bundle_cib_unmanaged_primitive_op_disabled = """
    <resources>
        <bundle id="A-bundle">
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </bundle>
    </resources>
"""
fixture_bundle_cib_unmanaged_both_op_disabled = """
    <resources>
        <bundle id="A-bundle">
            <meta_attributes id="A-bundle-meta_attributes">
                <nvpair id="A-bundle-meta_attributes-is-managed"
                    name="is-managed" value="false" />
            </meta_attributes>
            <docker image="pcs:test" />
            <primitive id="A" class="ocf" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
                <operations>
                    <op id="A-start" name="start" />
                    <op id="A-stop" name="stop" />
                    <op id="A-monitor" name="monitor" enabled="false"/>
                </operations>
            </primitive>
        </bundle>
    </resources>
"""

def fixture_report_no_monitors(resource):
    return (
        severities.WARNING,
        report_codes.RESOURCE_MANAGED_NO_MONITOR_ENABLED,
        {
            "resource_id": resource,
        },
        None
    )

class UnmanagePrimitive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_nonexistent_resource(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_managed)
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.unmanage(self.env_assist.get_env(), ["B"]),
            [
                fixture.report_not_found("B", "resources")
            ],
            expected_in_processor=False
        )

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_managed)
            .env.push_cib(resources=fixture_primitive_cib_unmanaged)
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_unmanaged)
            .env.push_cib(resources=fixture_primitive_cib_unmanaged)
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])


class ManagePrimitive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_nonexistent_resource(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_unmanaged)
        )

        self.env_assist.assert_raise_library_error(
            lambda: resource.manage(self.env_assist.get_env(), ["B"]),
            [
                fixture.report_not_found("B", "resources")
            ],
            expected_in_processor=False
        )

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_unmanaged)
            .env.push_cib(resources=fixture_primitive_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_managed(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_managed)
            .env.push_cib(resources=fixture_primitive_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])


class UnmanageGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_group_cib_managed)
            .env.push_cib(resources=fixture_group_cib_unmanaged_resource)
        )
        resource.unmanage(self.env_assist.get_env(), ["A1"])

    def test_group(self):
        (self.config
            .runner.cib.load(resources=fixture_group_cib_managed)
            .env.push_cib(
                resources=fixture_group_cib_unmanaged_all_resources
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])


class ManageGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(
                resources=fixture_group_cib_unmanaged_all_resources
            )
            .env.push_cib(resources=fixture_group_cib_unmanaged_resource)
        )
        resource.manage(self.env_assist.get_env(), ["A2"])

    def test_primitive_unmanaged_group(self):
        (self.config
            .runner.cib.load(
                resources=fixture_group_cib_unmanaged_resource_and_group
            )
            .env.push_cib(resources=fixture_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A1"])

    def test_group(self):
        (self.config
            .runner.cib.load(
                resources=fixture_group_cib_unmanaged_all_resources
            )
            .env.push_cib(resources=fixture_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_group_unmanaged_group(self):
        (self.config
            .runner.cib.load(
                resources=fixture_group_cib_unmanaged_resource_and_group
            )
            .env.push_cib(resources=fixture_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])


class UnmanageClone(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_managed)
            .env.push_cib(resources=fixture_clone_cib_unmanaged_primitive)
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_managed)
            .env.push_cib(resources=fixture_clone_cib_unmanaged_primitive)
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"])


class ManageClone(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_unmanaged_clone)
            .env.push_cib(resources=fixture_clone_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_unmanaged_primitive)
            .env.push_cib(resources=fixture_clone_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_both(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_unmanaged_both)
            .env.push_cib(resources=fixture_clone_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_unmanaged_clone)
            .env.push_cib(resources=fixture_clone_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])

    def test_clone_unmanaged_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_unmanaged_primitive)
            .env.push_cib(resources=fixture_clone_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])

    def test_clone_unmanaged_both(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_unmanaged_both)
            .env.push_cib(resources=fixture_clone_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])


class UnmanageMaster(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_managed)
            .env.push_cib(resources=fixture_master_cib_unmanaged_primitive)
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_master(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_managed)
            .env.push_cib(resources=fixture_master_cib_unmanaged_primitive)
        )
        resource.unmanage(self.env_assist.get_env(), ["A-master"])


class ManageMaster(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_unmanaged_primitive)
            .env.push_cib(resources=fixture_master_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_master(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_unmanaged_master)
            .env.push_cib(resources=fixture_master_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_both(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_unmanaged_both)
            .env.push_cib(resources=fixture_master_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_master(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_unmanaged_master)
            .env.push_cib(resources=fixture_master_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-master"])

    def test_master_unmanaged_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_unmanaged_primitive)
            .env.push_cib(resources=fixture_master_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-master"])

    def test_master_unmanaged_both(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_unmanaged_both)
            .env.push_cib(resources=fixture_master_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-master"])


class UnmanageClonedGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_group_cib_managed)
            .env.push_cib(
                resources=fixture_clone_group_cib_unmanaged_primitive
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A1"])

    def test_group(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_group_cib_managed)
            .env.push_cib(
                resources=fixture_clone_group_cib_unmanaged_all_primitives
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_group_cib_managed)
            .env.push_cib(
                resources=fixture_clone_group_cib_unmanaged_all_primitives
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"])


class ManageClonedGroup(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_unmanaged_primitive
            )
            .env.push_cib(resources=fixture_clone_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A1"])

    def test_primitive_unmanaged_all(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_unmanaged_everything
            )
            .env.push_cib(
                resources=fixture_clone_group_cib_unmanaged_primitive
            )
        )
        resource.manage(self.env_assist.get_env(), ["A2"])

    def test_group(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_unmanaged_all_primitives
            )
            .env.push_cib(resources=fixture_clone_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_group_unmanaged_all(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_unmanaged_everything
            )
            .env.push_cib(resources=fixture_clone_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_clone(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_group_cib_unmanaged_clone)
            .env.push_cib(resources=fixture_clone_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])

    def test_clone_unmanaged_all(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_unmanaged_everything
            )
            .env.push_cib(resources=fixture_clone_group_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-clone"])


class UnmanageBundle(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_managed)
            .env.push_cib(resources=fixture_bundle_cib_unmanaged_primitive)
        )
        resource.unmanage(self.env_assist.get_env(), ["A"])

    def test_bundle(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_managed)
            .env.push_cib(resources=fixture_bundle_cib_unmanaged_both)
        )
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_empty(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_empty_cib_managed)
            .env.push_cib(
                resources=fixture_bundle_empty_cib_unmanaged_bundle
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"])


class ManageBundle(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_unmanaged_primitive)
            .env.push_cib(resources=fixture_bundle_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_bundle(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_unmanaged_bundle)
            .env.push_cib(resources=fixture_bundle_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_primitive_unmanaged_both(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_unmanaged_both)
            .env.push_cib(resources=fixture_bundle_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"])

    def test_bundle(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_unmanaged_bundle)
            .env.push_cib(resources=fixture_bundle_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_unmanaged_primitive(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_unmanaged_primitive)
            .env.push_cib(resources=fixture_bundle_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_unmanaged_both(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_unmanaged_both)
            .env.push_cib(resources=fixture_bundle_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])

    def test_bundle_empty(self):
        (self.config
            .runner.cib.load(
                resources=fixture_bundle_empty_cib_unmanaged_bundle
            )
            .env.push_cib(resources=fixture_bundle_empty_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A-bundle"])


class MoreResources(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    fixture_cib_managed = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
            </primitive>
            <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
            </primitive>
        </resources>
    """
    fixture_cib_unmanaged = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                <meta_attributes id="A-meta_attributes">
                    <nvpair id="A-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
            <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                <meta_attributes id="B-meta_attributes">
                    <nvpair id="B-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
            <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                <meta_attributes id="C-meta_attributes">
                    <nvpair id="C-meta_attributes-is-managed"
                        name="is-managed" value="false" />
                </meta_attributes>
            </primitive>
        </resources>
    """

    def test_success_unmanage(self):
        fixture_cib_unmanaged = """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                </primitive>
                <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                    <meta_attributes id="C-meta_attributes">
                        <nvpair id="C-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
            </resources>
        """
        (self.config
            .runner.cib.load(resources=self.fixture_cib_managed)
            .env.push_cib(resources=fixture_cib_unmanaged)
        )
        resource.unmanage(self.env_assist.get_env(), ["A", "C"])

    def test_success_manage(self):
        fixture_cib_managed = """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes">
                        <nvpair id="B-meta_attributes-is-managed"
                            name="is-managed" value="false" />
                    </meta_attributes>
                </primitive>
                <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                </primitive>
            </resources>
        """
        (self.config
            .runner.cib.load(resources=self.fixture_cib_unmanaged)
            .env.push_cib(resources=fixture_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A", "C"])

    def test_bad_resource_unmanage(self):
        (self.config
            .runner.cib.load(resources=self.fixture_cib_managed)
        )

        self.env_assist.assert_raise_library_error(
            lambda:
            resource.unmanage(self.env_assist.get_env(), ["B", "X", "Y", "A"]),
            [
                fixture.report_not_found("X", "resources"),
                fixture.report_not_found("Y", "resources"),
            ],
            expected_in_processor=False
        )

    def test_bad_resource_enable(self):
        (self.config
            .runner.cib.load(resources=self.fixture_cib_unmanaged)
        )

        self.env_assist.assert_raise_library_error(
            lambda:
            resource.manage(self.env_assist.get_env(), ["B", "X", "Y", "A"]),
            [
                fixture.report_not_found("X", "resources"),
                fixture.report_not_found("Y", "resources"),
            ],
            expected_in_processor=False
        )


class WithMonitor(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_unmanage_noop(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_managed)
            .env.push_cib(resources=fixture_primitive_cib_unmanaged)
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_manage_noop(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_unmanaged)
            .env.push_cib(resources=fixture_primitive_cib_managed)
        )
        resource.manage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_primitive_cib_unmanaged_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_manage(self):
        (self.config
            .runner.cib.load(
                resources=fixture_primitive_cib_unmanaged_op_disabled
            )
            .env.push_cib(resources=fixture_primitive_cib_managed_op_enabled)
        )
        resource.manage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_enabled_monitors(self):
        (self.config
            .runner.cib.load(resources=fixture_primitive_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_primitive_cib_unmanaged_op_enabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], False)

    def test_manage_disabled_monitors(self):
        (self.config
            .runner.cib.load(
                resources=fixture_primitive_cib_unmanaged_op_disabled
            )
            .env.push_cib(
                resources=fixture_primitive_cib_managed_op_disabled
            )
        )
        resource.manage(self.env_assist.get_env(), ["A"], False)
        self.env_assist.assert_reports([
            fixture_report_no_monitors("A"),
        ])

    def test_unmanage_clone(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_clone_cib_unmanaged_primitive_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"], True)

    def test_unmanage_in_clone(self):
        (self.config
            .runner.cib.load(resources=fixture_clone_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_clone_cib_unmanaged_primitive_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_master(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_master_cib_unmanaged_primitive_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A-master"], True)

    def test_unmanage_in_master(self):
        (self.config
            .runner.cib.load(resources=fixture_master_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_master_cib_unmanaged_primitive_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_clone_with_group(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_managed_op_enabled
            )
            .env.push_cib(resources=
                fixture_clone_group_cib_unmanaged_all_primitives_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A-clone"], True)

    def test_unmanage_group_in_clone(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_managed_op_enabled
            )
            .env.push_cib(resources=
                fixture_clone_group_cib_unmanaged_all_primitives_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_in_cloned_group(self):
        (self.config
            .runner.cib.load(
                resources=fixture_clone_group_cib_managed_op_enabled
            )
            .env.push_cib(resources=
                fixture_clone_group_cib_unmanaged_primitive_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A1"], True)

    def test_unmanage_bundle(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_bundle_cib_unmanaged_both_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"], True)

    def test_unmanage_in_bundle(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_cib_managed_op_enabled)
            .env.push_cib(
                resources=fixture_bundle_cib_unmanaged_primitive_op_disabled
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A"], True)

    def test_unmanage_bundle_empty(self):
        (self.config
            .runner.cib.load(resources=fixture_bundle_empty_cib_managed)
            .env.push_cib(
                resources=fixture_bundle_empty_cib_unmanaged_bundle
            )
        )
        resource.unmanage(self.env_assist.get_env(), ["A-bundle"], True)



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.test.tools.pcs_unittest import TestCase

from pcs.cli.booth.console_report import CODE_TO_MESSAGE_BUILDER_MAP
from pcs.common import report_codes as codes


class BoothConfigAccetedByNodeTest(TestCase):
    def setUp(self):
        self.build = CODE_TO_MESSAGE_BUILDER_MAP[
            codes.BOOTH_CONFIG_ACCEPTED_BY_NODE
        ]

    def test_crete_message_with_empty_name_list(self):
        self.assertEqual("Booth config saved.", self.build({
            "node": None,
            "name_list": [],
        }))

    def test_crete_message_with_name_booth_only(self):
        self.assertEqual("Booth config saved.", self.build({
            "node": None,
            "name_list": ["booth"],
        }))

    def test_crete_message_with_single_name(self):
        self.assertEqual("Booth config(s) (some) saved.", self.build({
            "node": None,
            "name_list": ["some"],
        }))

    def test_crete_message_with_multiple_name(self):
        self.assertEqual("Booth config(s) (some, another) saved.", self.build({
            "node": None,
            "name_list": ["some", "another"],
        }))

    def test_crete_message_with_empty_node(self):
        self.assertEqual(
            "node1: Booth config(s) (some, another) saved.",
            self.build({
                "node": "node1",
                "name_list": ["some", "another"],
            }),
        )

class BoothConfigDistributionNodeErrorTest(TestCase):
    def setUp(self):
        self.build = CODE_TO_MESSAGE_BUILDER_MAP[
            codes.BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR
        ]

    def test_create_message_for_empty_name(self):
        self.assertEqual(
            "Unable to save booth config on node 'node1': reason1",
            self.build({
                "node": "node1",
                "reason": "reason1",
                "name": None,
            })
        )

    def test_create_message_for_booth_name(self):
        self.assertEqual(
            "Unable to save booth config on node 'node1': reason1",
            self.build({
                "node": "node1",
                "reason": "reason1",
                "name": "booth",
            })
        )

    def test_create_message_for_another_name(self):
        self.assertEqual(
            "Unable to save booth config (another) on node 'node1': reason1",
            self.build({
                "node": "node1",
                "reason": "reason1",
                "name": "another",
            })
        )

class BoothConfigReadErrorTest(TestCase):
    def setUp(self):
        self.build = CODE_TO_MESSAGE_BUILDER_MAP[
            codes.BOOTH_CONFIG_READ_ERROR
        ]

    def test_create_message_for_empty_name(self):
        self.assertEqual("Unable to read booth config.", self.build({
            "name": None,
        }))

    def test_create_message_for_booth_name(self):
        self.assertEqual("Unable to read booth config.", self.build({
            "name": "booth",
        }))

    def test_create_message_for_another_name(self):
        self.assertEqual("Unable to read booth config (another).", self.build({
            "name": "another",
        }))

class BoothFetchingConfigFromNodeTest(TestCase):
    def setUp(self):
        self.build = CODE_TO_MESSAGE_BUILDER_MAP[
            codes.BOOTH_FETCHING_CONFIG_FROM_NODE
        ]

    def test_create_message_for_empty_name(self):
        self.assertEqual(
            "Fetching booth config from node 'node1'...",
            self.build({
                "config": None,
                "node": "node1",
            })
        )

    def test_create_message_for_booth_name(self):
        self.assertEqual(
            "Fetching booth config from node 'node1'...",
            self.build({
                "config": "booth",
                "node": "node1",
            })
        )

    def test_create_message_for_another_name(self):
        self.assertEqual(
            "Fetching booth config 'another' from node 'node1'...",
            self.build({
                "config": "another",
                "node": "node1",
            })
        )



from __future__ import (
    absolute_import,
    division,
    print_function,
)

import shutil
from textwrap import dedent

from pcs.test.tools.assertions import AssertPcsMixin
from pcs.test.tools.misc import get_test_resource as rc
from pcs.test.tools.pcs_runner import PcsRunner
from pcs.test.tools.pcs_unittest import TestCase


class StonithWarningTest(TestCase, AssertPcsMixin):
    empty_cib = rc("cib-empty.xml")
    temp_cib = rc("temp-cib.xml")

    def setUp(self):
        shutil.copy(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib)

    def fixture_stonith_action(self):
        self.assert_pcs_success(
            "stonith create Sa fence_apc ipaddr=i login=l action=reboot --force",
            "Warning: stonith option 'action' is deprecated and should not be"
                " used, use pcmk_off_action, pcmk_reboot_action instead\n"
        )

    def fixture_stonith_cycle(self):
        self.assert_pcs_success(
            "stonith create Sc fence_ipmilan method=cycle"
        )

    def fixture_resource(self):
        self.assert_pcs_success(
            "resource create dummy ocf:pacemaker:Dummy action=reboot "
                "method=cycle --force"
            ,
            "Warning: invalid resource options: 'action', 'method', allowed "
                "options are: envfile, fail_start_on, fake, op_sleep, passwd, "
                "state, trace_file, trace_ra\n"
        )

    def test_warning_stonith_action(self):
        self.fixture_stonith_action()
        self.fixture_resource()
        self.assert_pcs_success(
            "status",
            stdout_start=dedent("""\
                Cluster name: test99

                WARNINGS:
                Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: Sa

                Stack: unknown
                Current DC: NONE
            """)
        )

    def test_warning_stonith_method_cycle(self):
        self.fixture_stonith_cycle()
        self.fixture_resource()
        self.assert_pcs_success(
            "status",
            stdout_start=dedent("""\
                Cluster name: test99

                WARNINGS:
                Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': Sc

                Stack: unknown
                Current DC: NONE
            """)
        )

    def test_stonith_warnings(self):
        self.fixture_stonith_action()
        self.fixture_stonith_cycle()
        self.fixture_resource()
        self.assert_pcs_success(
            "status",
            stdout_start=dedent("""\
                Cluster name: test99

                WARNINGS:
                Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: Sa
                Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': Sc

                Stack: unknown
                Current DC: NONE
            """)
        )

    def test_warn_when_no_stonith(self):
        self.assert_pcs_success(
            "status",
            stdout_start=dedent("""\
                Cluster name: test99

                WARNINGS:
                No stonith devices and stonith-enabled is not false

                Stack: unknown
                Current DC: NONE
            """)
        )

    def test_disabled_stonith_does_not_care_about_missing_devices(self):
        self.assert_pcs_success("property set stonith-enabled=false")
        self.assert_pcs_success(
            "status",
            stdout_start=dedent("""\
                Cluster name: test99
                Stack: unknown
                Current DC: NONE
            """)
        )



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.lib import reports, validate
from pcs.lib.cib.tools import does_id_exist
from pcs.lib.cib.nvpair import(
    has_meta_attribute,
    arrange_first_meta_attributes,
    get_meta_attribute_value,
)
from pcs.lib.node import (
    NodeAddresses,
    node_addresses_contain_host,
    node_addresses_contain_name,
)
from pcs.lib.xml_tools import remove_when_pointless


#TODO pcs currently does not care about multiple meta_attributes and here
#we don't care as well
GUEST_OPTIONS = [
    'remote-port',
    'remote-addr',
    'remote-connect-timeout',
]

def validate_conflicts(tree, nodes, node_name, options):
    report_list = []
    if(
        does_id_exist(tree, node_name)
        or
        node_addresses_contain_name(nodes, node_name)
        or (
            "remote-addr" not in options
            and
            node_addresses_contain_host(nodes, node_name)
        )
    ):
        report_list.append(reports.id_already_exists(node_name))

    if(
        "remote-addr" in options
        and
        node_addresses_contain_host(nodes, options["remote-addr"])
    ):
        report_list.append(reports.id_already_exists(options["remote-addr"]))
    return report_list

def is_node_name_in_options(options):
    return "remote-node" in options

def get_guest_option_value(options, default=None):
    return options.get("remote-node", default)


def validate_set_as_guest(tree, nodes, node_name, options):
    report_list = validate.names_in(
        GUEST_OPTIONS,
        options.keys(),
        "guest",
    )

    validator_list = [
        validate.value_time_interval("remote-connect-timeout"),
        validate.value_port_number("remote-port"),
    ]

    report_list.extend(
        validate.run_collection_of_option_validators(options, validator_list)
    )

    report_list.extend(
        validate_conflicts(tree, nodes, node_name, options)
    )

    if not node_name.strip():
        report_list.append(
            reports.invalid_option_value(
                "node name",
                node_name,
                "no empty value",
            )
        )

    return report_list

def is_guest_node(resource_element):
    """
    Return True if resource_element is already set as guest node.

    etree.Element resource_element is a search element
    """
    return has_meta_attribute(resource_element, "remote-node")

def validate_is_not_guest(resource_element):
    """
    etree.Element resource_element
    """
    if not is_guest_node(resource_element):
        return []

    return [
        reports.resource_is_guest_node_already(
            resource_element.attrib["id"]
        )
    ]

def set_as_guest(
    resource_element, node, addr=None, port=None, connect_timeout=None
):
    """
    Set resource as guest node.

    etree.Element resource_element

    """
    meta_options = {"remote-node": str(node)}
    if addr:
        meta_options["remote-addr"] = str(addr)
    if port:
        meta_options["remote-port"] = str(port)
    if connect_timeout:
        meta_options["remote-connect-timeout"] = str(connect_timeout)

    arrange_first_meta_attributes(resource_element, meta_options)

def unset_guest(resource_element):
    """
    Unset resource as guest node.

    etree.Element resource_element
    """
    guest_nvpair_list = resource_element.xpath(
        "./meta_attributes/nvpair[{0}]".format(
            " or ".join([
                '@name="{0}"'.format(option)
                for option in (GUEST_OPTIONS + ["remote-node"])
            ])
        )
    )
    for nvpair in guest_nvpair_list:
        meta_attributes = nvpair.getparent()
        meta_attributes.remove(nvpair)
        remove_when_pointless(meta_attributes)

def get_node(meta_attributes):
    """
    Return NodeAddresses with corresponding to guest node in meta_attributes.
    Return None if meta_attributes does not mean guest node

    etree.Element meta_attributes is a researched element
    """
    host = None
    name = None
    for nvpair in meta_attributes:
        if nvpair.attrib.get("name", "") == "remote-addr":
            host = nvpair.attrib["value"]
        if nvpair.attrib.get("name", "") == "remote-node":
            name = nvpair.attrib["value"]
            if host is None:
                host = name
    return NodeAddresses(host, name=name) if name else None

def get_host_from_options(node_name, meta_options):
    """
    Return host from node_name meta options.
    dict meta_options
    """
    return meta_options.get("remote-addr", node_name)

def get_node_name_from_options(meta_options, default=None):
    """
    Return node_name from meta options.
    dict meta_options
    """
    return meta_options.get("remote-node", default)


def get_host(resource_element):
    host = get_meta_attribute_value(resource_element, "remote-addr")
    if host:
        return host

    return get_meta_attribute_value(resource_element, "remote-node")

def find_node_list(resources_section):
    """
    Return list of nodes from resources_section

    etree.Element resources_section is a researched element
    """
    return [
        get_node(meta_attrs) for meta_attrs in resources_section.xpath("""
            .//primitive
                /meta_attributes[
                    nvpair[
                        @name="remote-node"
                        and
                        string-length(@value) > 0
                    ]
                ]
        """)
    ]

def find_node_resources(resources_section, node_identifier):
    """
    Return list of etree.Eleent primitives that are guest nodes.

    etree.Element resources_section is a researched element
    string node_identifier could be id of resource, node name or node address
    """
    resources = resources_section.xpath("""
        .//primitive[
            (
                @id="{0}"
                and
                meta_attributes[
                    nvpair[
                        @name="remote-node"
                        and
                        string-length(@value) > 0
                    ]
                ]
            )
            or
            meta_attributes[
                nvpair[
                    (
                        @name="remote-addr"
                        or
                        @name="remote-node"
                    )
                    and
                    @value="{0}"
                ]
            ]
        ]
    """.format(node_identifier))
    return resources



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from functools import partial

from lxml import etree

from pcs.test.tools.pcs_unittest import TestCase
from pcs.test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
)
from pcs.test.tools import fixture
from pcs.test.tools.misc import get_test_resource as rc
from pcs.test.tools.pcs_unittest import mock
from pcs.test.tools.xml import get_xml_manipulation_creator_from_file

from pcs.common import report_codes
from pcs.common.tools import Version
from pcs.lib.errors import ReportItemSeverity as severities

from pcs.lib.cib import tools as lib

class CibToolsTest(TestCase):
    def setUp(self):
        self.create_cib = get_xml_manipulation_creator_from_file(
            rc("cib-empty.xml")
        )
        self.cib = self.create_cib()

    def fixture_add_primitive_with_id(self, element_id):
        self.cib.append_to_first_tag_name(
            "resources",
            '<primitive id="{0}" class="ocf" provider="heartbeat" type="Dummy"/>'
                .format(element_id)
        )


class IdProviderTest(CibToolsTest):
    def setUp(self):
        super(IdProviderTest, self).setUp()
        self.provider = lib.IdProvider(self.cib.tree)

    def fixture_report(self, id):
        return (
            severities.ERROR,
            report_codes.ID_ALREADY_EXISTS,
            {
                "id": id,
            },
            None
        )


class IdProviderBook(IdProviderTest):
    def test_nonexisting_id(self):
        assert_report_item_list_equal(
            self.provider.book_ids("myId"),
            []
        )

    def test_existing_id(self):
        self.fixture_add_primitive_with_id("myId")
        assert_report_item_list_equal(
            self.provider.book_ids("myId"),
            [
                self.fixture_report("myId"),
            ]
        )

    def test_double_book(self):
        assert_report_item_list_equal(
            self.provider.book_ids("myId"),
            []
        )
        assert_report_item_list_equal(
            self.provider.book_ids("myId"),
            [
                self.fixture_report("myId"),
            ]
        )

    def test_more_ids(self):
        assert_report_item_list_equal(
            self.provider.book_ids("myId1", "myId2"),
            []
        )
        assert_report_item_list_equal(
            self.provider.book_ids("myId1", "myId2"),
            [
                self.fixture_report("myId1"),
                self.fixture_report("myId2"),
            ]
        )

    def test_complex(self):
        # test ids existing in the cib, double booked, available
        # test reports not repeated
        self.fixture_add_primitive_with_id("myId1")
        self.fixture_add_primitive_with_id("myId2")
        assert_report_item_list_equal(
            self.provider.book_ids(
                "myId1", "myId2", "myId3", "myId2", "myId3", "myId4", "myId3"
            ),
            [
                self.fixture_report("myId1"),
                self.fixture_report("myId2"),
                self.fixture_report("myId3"),
            ]
        )


class IdProviderAllocate(IdProviderTest):
    def test_nonexisting_id(self):
        self.assertEqual("myId",  self.provider.allocate_id("myId"))

    def test_existing_id(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertEqual("myId-1",  self.provider.allocate_id("myId"))

    def test_allocate_books(self):
        self.assertEqual("myId",  self.provider.allocate_id("myId"))
        self.assertEqual("myId-1",  self.provider.allocate_id("myId"))

    def test_booked_ids(self):
        self.fixture_add_primitive_with_id("myId")
        assert_report_item_list_equal(
            self.provider.book_ids("myId-1"),
            []
        )
        self.assertEqual("myId-2",  self.provider.allocate_id("myId"))


class DoesIdExistTest(CibToolsTest):
    def test_existing_id(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertTrue(lib.does_id_exist(self.cib.tree, "myId"))

    def test_nonexisting_id(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertFalse(lib.does_id_exist(self.cib.tree, "otherId"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "myid"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, " myId"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "myId "))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "my Id"))

    def test_ignore_status_section(self):
        self.cib.append_to_first_tag_name("status", """
            <elem1 id="status-1">
                <elem1a id="status-1a">
                    <elem1aa id="status-1aa"/>
                    <elem1ab id="status-1ab"/>
                </elem1a>
                <elem1b id="status-1b">
                    <elem1ba id="status-1ba"/>
                    <elem1bb id="status-1bb"/>
                </elem1b>
            </elem1>
        """)
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1a"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1aa"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1ab"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1b"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1ba"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1bb"))

    def test_ignore_acl_target(self):
        self.cib.append_to_first_tag_name(
            "configuration",
            """
            <acls>
                <acl_target id="target1"/>
            </acls>
            """
        )
        self.assertFalse(lib.does_id_exist(self.cib.tree, "target1"))

    def test_ignore_acl_role_references(self):
        self.cib.append_to_first_tag_name(
            "configuration",
            """
            <acls>
                <acl_target id="target1">
                    <role id="role1"/>
                    <role id="role2"/>
                </acl_target>
            </acls>
            """
        )
        self.assertFalse(lib.does_id_exist(self.cib.tree, "role1"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "role2"))

    def test_ignore_sections_directly_under_cib(self):
        #this is side effect of current implementation but is not problem since
        #id attribute is not allowed for elements directly under cib
        tree = etree.fromstring('<cib><direct id="a"/></cib>')
        self.assertFalse(lib.does_id_exist(tree, "a"))

    def test_find_id_when_cib_is_not_root_element(self):
        #for example we have only part of xml
        tree = etree.fromstring('<root><direct id="a"/></root>')
        self.assertTrue(lib.does_id_exist(tree, "a"))

    def test_find_remote_node_pacemaker_internal_id(self):
        tree = etree.fromstring("""
            <cib>
                <configuration>
                    <resources>
                        <primitive id="b">
                            <meta_attributes>
                                <nvpair name="remote-node" value="a"/>
                            </meta_attributes>
                        </primitive>
                    </resources>
                </configuration>
            </cib>
        """)
        self.assertTrue(lib.does_id_exist(tree, "a"))

class FindUniqueIdTest(CibToolsTest):
    def test_already_unique(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertEqual("other", lib.find_unique_id(self.cib.tree, "other"))

    def test_add_suffix(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertEqual("myId-1", lib.find_unique_id(self.cib.tree, "myId"))

        self.fixture_add_primitive_with_id("myId-1")
        self.assertEqual("myId-2", lib.find_unique_id(self.cib.tree, "myId"))

    def test_suffix_not_needed(self):
        self.fixture_add_primitive_with_id("myId-1")
        self.assertEqual("myId", lib.find_unique_id(self.cib.tree, "myId"))

    def test_add_first_available_suffix(self):
        self.fixture_add_primitive_with_id("myId")
        self.fixture_add_primitive_with_id("myId-1")
        self.fixture_add_primitive_with_id("myId-3")
        self.assertEqual("myId-2", lib.find_unique_id(self.cib.tree, "myId"))

    def test_reserved_ids(self):
        self.fixture_add_primitive_with_id("myId-1")
        self.assertEqual(
            "myId-3",
            lib.find_unique_id(self.cib.tree, "myId", ["myId", "myId-2"])
        )

class CreateNvsetIdTest(TestCase):
    def test_create_plain_id_when_no_confilicting_id_there(self):
        context = etree.fromstring('<cib><a id="b"/></cib>')
        self.assertEqual(
            "b-name",
            lib.create_subelement_id(context.find(".//a"), "name")
        )

    def test_create_decorated_id_when_conflicting_id_there(self):
        context = etree.fromstring(
            '<cib><a id="b"><c id="b-name"/></a></cib>'
        )
        self.assertEqual(
            "b-name-1",
            lib.create_subelement_id(context.find(".//a"), "name")
        )

class GetConfigurationTest(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual(
            "configuration",
            lib.get_configuration(self.cib.tree).tag
        )

    def test_raise_if_missing(self):
        for conf in self.cib.tree.findall(".//configuration"):
            conf.getparent().remove(conf)
        assert_raise_library_error(
            lambda: lib.get_configuration(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration",
                }
            ),
        )

class GetConstraintsTest(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual(
            "constraints",
            lib.get_constraints(self.cib.tree).tag
        )

    def test_raise_if_missing(self):
        for section in self.cib.tree.findall(".//configuration/constraints"):
            section.getparent().remove(section)
        assert_raise_library_error(
            lambda: lib.get_constraints(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration/constraints",
                }
            ),
        )

class GetResourcesTest(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual(
            "resources",
            lib.get_resources(self.cib.tree).tag
        )

    def test_raise_if_missing(self):
        for section in self.cib.tree.findall(".//configuration/resources"):
            section.getparent().remove(section)
        assert_raise_library_error(
            lambda: lib.get_resources(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration/resources",
                }
            ),
        )

class GetNodes(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual(
            "nodes",
            lib.get_nodes(self.cib.tree).tag
        )

    def test_raise_if_missing(self):
        for section in self.cib.tree.findall(".//configuration/nodes"):
            section.getparent().remove(section)
        assert_raise_library_error(
            lambda: lib.get_nodes(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration/nodes",
                },
                None
            ),
        )

class GetAclsTest(CibToolsTest):
    def test_success_if_exists(self):
        self.cib.append_to_first_tag_name(
            "configuration",
            '<acls><acl_role id="test_role" /></acls>'
        )
        self.assertEqual(
            "test_role",
            lib.get_acls(self.cib.tree)[0].get("id")
        )

    def test_success_if_missing(self):
        acls = lib.get_acls(self.cib.tree)
        self.assertEqual("acls", acls.tag)
        self.assertEqual("configuration", acls.getparent().tag)

class GetFencingTopology(CibToolsTest):
    def test_success_if_exists(self):
        self.cib.append_to_first_tag_name(
            "configuration",
            "<fencing-topology />"
        )
        self.assertEqual(
            "fencing-topology",
            lib.get_fencing_topology(self.cib.tree).tag
        )

    def test_success_if_missing(self):
        ft = lib.get_fencing_topology(self.cib.tree)
        self.assertEqual("fencing-topology", ft.tag)
        self.assertEqual("configuration", ft.getparent().tag)


@mock.patch('pcs.lib.cib.tools.does_id_exist')
class ValidateIdDoesNotExistsTest(TestCase):
    def test_success_when_id_does_not_exists(self, does_id_exists):
        does_id_exists.return_value = False
        lib.validate_id_does_not_exist("tree", "some-id")
        does_id_exists.assert_called_once_with("tree", "some-id")

    def test_raises_whne_id_exists(self, does_id_exists):
        does_id_exists.return_value = True
        assert_raise_library_error(
            lambda: lib.validate_id_does_not_exist("tree", "some-id"),
            (
                severities.ERROR,
                report_codes.ID_ALREADY_EXISTS,
                {"id": "some-id"},
            ),
        )
        does_id_exists.assert_called_once_with("tree", "some-id")


class GetPacemakerVersionByWhichCibWasValidatedTest(TestCase):
    def test_missing_attribute(self):
        assert_raise_library_error(
            lambda: lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML("<cib/>")
            ),
            (
                severities.ERROR,
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                {
                    "reason": "the attribute 'validate-with' of the element"
                        " 'cib' is missing"
                }
            )
        )

    def test_invalid_version(self):
        assert_raise_library_error(
            lambda: lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML('<cib validate-with="something-1.2.3"/>')
            ),
            (
                severities.ERROR,
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                {
                    "reason": "the attribute 'validate-with' of the element"
                        " 'cib' has an invalid value: 'something-1.2.3'"
                }
            )
        )

    def test_no_revision(self):
        self.assertEqual(
            Version(1, 2),
            lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML('<cib validate-with="pacemaker-1.2"/>')
            )
        )

    def test_with_revision(self):
        self.assertEqual(
            Version(1, 2, 3),
            lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML('<cib validate-with="pacemaker-1.2.3"/>')
            )
        )


class getCibCrmFeatureSet(TestCase):
    def test_success(self):
        self.assertEqual(
            Version(3, 0, 9),
            lib.get_cib_crm_feature_set(
                etree.XML('<cib crm_feature_set="3.0.9" />')
            )
        )

    def test_success_no_revision(self):
        self.assertEqual(
            Version(3, 1),
            lib.get_cib_crm_feature_set(
                etree.XML('<cib crm_feature_set="3.1" />')
            )
        )

    def test_missing_attribute(self):
        assert_raise_library_error(
            lambda: lib.get_cib_crm_feature_set(
                etree.XML("<cib />")
            ),
            fixture.error(
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                reason=(
                    "the attribute 'crm_feature_set' of the element 'cib' is "
                    "missing"
                )
            )
        )

    def test_missing_attribute_none(self):
        self.assertEqual(
            None,
            lib.get_cib_crm_feature_set(
                etree.XML('<cib />'),
                none_if_missing=True
            )
        )

    def test_invalid_version(self):
        assert_raise_library_error(
            lambda: lib.get_cib_crm_feature_set(
                etree.XML('<cib crm_feature_set="3" />')
            ),
            fixture.error(
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                reason=(
                    "the attribute 'crm_feature_set' of the element 'cib' has "
                    "an invalid value: '3'"
                )
            )
        )


find_group = partial(lib.find_element_by_tag_and_id, "group")
class FindTagWithId(TestCase):
    def test_returns_element_when_exists(self):
        tree = etree.fromstring(
            '<cib><resources><group id="a"/></resources></cib>'
        )
        element = find_group(tree.find(".//resources"), "a")
        self.assertEqual("group", element.tag)
        self.assertEqual("a", element.attrib["id"])

    def test_returns_element_when_exists_one_of_tags(self):
        tree = etree.fromstring("""
            <cib>
                <resources>
                    <group id="a"/>
                    <primitive id="b"/>
                </resources>
            </cib>
        """)
        element = lib.find_element_by_tag_and_id(
            ["group", "primitive"],
            tree.find(".//resources"),
            "a"
        )
        self.assertEqual("group", element.tag)
        self.assertEqual("a", element.attrib["id"])

    def test_raises_when_is_under_another_tag(self):
        tree = etree.fromstring(
            '<cib><resources><primitive id="a"/></resources></cib>'
        )

        assert_raise_library_error(
            lambda: find_group(tree.find(".//resources"), "a"),
            (
                severities.ERROR,
                report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                {
                    "id": "a",
                    "expected_types": ["group"],
                    "current_type": "primitive",
                },
            ),
        )

    def test_raises_when_is_under_another_context(self):
        tree = etree.fromstring("""
            <cib>
                <resources>
                    <group id="g1"><primitive id="a"/></group>
                    <group id="g2"><primitive id="b"/></group>
                </resources>
            </cib>
        """)
        assert_raise_library_error(
            lambda: lib.find_element_by_tag_and_id(
                "primitive",
                tree.find('.//resources/group[@id="g2"]'),
                "a"
            ),
            (
                severities.ERROR,
                report_codes.OBJECT_WITH_ID_IN_UNEXPECTED_CONTEXT,
                {
                    "type": "primitive",
                    "id": "a",
                    "expected_context_type": "group",
                    "expected_context_id": "g2",
                },
            ),
        )

    def test_raises_when_id_does_not_exists(self):
        tree = etree.fromstring('<cib><resources/></cib>')
        assert_raise_library_error(
            lambda: find_group(tree.find('.//resources'), "a"),
            (
                severities.ERROR,
                report_codes.ID_NOT_FOUND,
                {
                    "id": "a",
                    "expected_types": ["group"],
                    "context_type": "resources",
                    "context_id": "",
                },
                None
            ),
        )
        assert_raise_library_error(
            lambda: find_group(
                tree.find('.//resources'),
                "a",
                id_types=["resource group"]
            ),
            (
                severities.ERROR,
                report_codes.ID_NOT_FOUND,
                {
                    "id": "a",
                    "expected_types": ["resource group"],
                    "context_type": "resources",
                    "context_id": "",
                },
                None
            ),
        )

    def test_returns_none_if_id_do_not_exists(self):
        tree = etree.fromstring('<cib><resources/></cib>')
        self.assertIsNone(find_group(
            tree.find('.//resources'),
            "a",
            none_if_id_unused=True
        ))



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from lxml import etree

from pcs.lib import reports
from pcs.lib.cib.tools import find_element_by_tag_and_id
from pcs.lib.errors import LibraryError


TAG = "group"

def is_group(resource_el):
    return resource_el.tag == TAG

def provide_group(resources_section, group_id):
    """
    Provide group with id=group_id. Create new group if group with id=group_id
    does not exists.

    etree.Element resources_section is place where new group will be appended
    string group_id is id of group
    """
    group_element = find_element_by_tag_and_id(
        TAG,
        resources_section,
        group_id,
        none_if_id_unused=True
    )
    if group_element is None:
        group_element = etree.SubElement(resources_section, TAG, id=group_id)
    return group_element

def place_resource(
    group_element, primitive_element,
    adjacent_resource_id=None, put_after_adjacent=False
):
    """
    Add resource into group. This function is also applicable for a modification
    of the resource position because the primitive element is replanted from
    anywhere (including group itself) to concrete place inside group.

    etree.Element group_element is element where to put primitive_element
    etree.Element primitive_element is element for placement
    string adjacent_resource_id is id of the existing resource in group.
        primitive_element will be put beside adjacent_resource_id if specified.
    bool put_after_adjacent is flag where put primitive_element:
        before adjacent_resource_id if put_after_adjacent=False
        after adjacent_resource_id if put_after_adjacent=True
        Note that it make sense only if adjacent_resource_id is specified
    """
    if primitive_element.attrib["id"] == adjacent_resource_id:
        raise LibraryError(reports.resource_cannot_be_next_to_itself_in_group(
            adjacent_resource_id,
            group_element.attrib["id"],
        ))

    if not adjacent_resource_id:
        return group_element.append(primitive_element)

    adjacent_resource = find_element_by_tag_and_id(
        "primitive",
        group_element,
        adjacent_resource_id,
    )

    if put_after_adjacent and adjacent_resource.getnext() is None:
        return group_element.append(primitive_element)

    index = group_element.index(
        adjacent_resource.getnext() if put_after_adjacent
        else adjacent_resource
    )
    group_element.insert(index, primitive_element)

def get_inner_resources(group_el):
    return group_el.xpath("./primitive")
